#include <_base64.h>
#include <SPI.h>
#include <Wire.h>
#include <WiFiUdp.h>
#include <WiFiServerSecure.h>
#include <WiFiServer.h>
#include <WiFiClientSecure.h>
#include <WiFiClient.h>
#include <ESP8266WiFiType.h>
#include <ESP8266WiFiSTA.h>
#include <ESP8266WiFiScan.h>
#include <ESP8266WiFiMulti.h>
#include <ESP8266WiFiGeneric.h>
#include <ESP8266WiFiAP.h>
#include <ESP8266WiFi.h>
#include <gfxfont.h>
#include <Adafruit_SPITFT_Macros.h>
#include <Adafruit_SPITFT.h>
#include <Adafruit_GFX.h>
#include <Adafruit_PCD8544.h>

// USER CONFIG HERE \/
//#define DEBUG
#define pinLED 16
#define pinBTN 2
#define WiFiSsid "ssid"
#define WiFiPass "heslo"
#define pop3Server "pop3.seznam.cz"
#define pop3User "user jozin@zbazin.cz"  // DO NOT REMOVE user FROM BEGINING OF STRING. For username jozin@zbazin.cz type there (with quotes) "user jozin@zbazin.cz"
#define pop3Password "pass 1234" // DO NOT REMOVE pass FROM BEGINING OF STRING. For password 1234 type there (with quotes) "pass 1234"
// USER CONFIG HERE /\

#define INIT_TIMEOUT int millisStart = millis();
#define CHECK_TIMEOUT_NOT_HAPPEND(DURATION) (millis() - millisStart < (DURATION))
#define CHECK_TIMEOUT_HAPPEND(DURATION, RETVAL_IF_HAPPEND) if (millis() - millisStart > (DURATION)) { /*Serial.println("TIMEOUT!");*/ return RETVAL_IF_HAPPEND; }

#define INIT_TIMEOUT2 int millisStart2 = millis();
#define CHECK_TIMEOUT_NOT_HAPPEND2(DURATION) (millis() - millisStart2 < (DURATION))
#define CHECK_TIMEOUT_HAPPEND2(DURATION, RETVAL_IF_HAPPEND) if (millis() - millisStart > (DURATION)) { /*Serial.println("TIMEOUT!");*/ return RETVAL_IF_HAPPEND; }


Adafruit_PCD8544 display = Adafruit_PCD8544(12, 5, 4);
WiFiClient tcpCon;
uint8_t buff[10000];

String lastDate;
String lastSubject;
String lastFrom;

void setup() {
	Serial.begin(9600);
	Serial.println("\r\nInitialization started");
	delay(500);
	display.begin();
	display.setContrast(65);
	display.display();

	WiFi.mode(WIFI_STA);
	WiFi.disconnect();
	WiFi.begin(WiFiSsid, WiFiPass);
	waitForConnection(3);

	pinMode(pinLED, OUTPUT);
	pinMode(pinBTN, INPUT);
	digitalWrite(pinLED, 0);
	attachInterrupt(pinBTN, stopLED, FALLING);
}

void displayFailure() {
	display.clearDisplay();
	display.display();
	display.println("\r\nError occured!");
	display.println("");
#ifndef DEBUG
	display.println("Enable DEBUG");
#endif
	display.println("See serial");
	display.display();

	while (1) {
		delay(1);
	}
}

void stopLED() {
	digitalWrite(pinLED, 0);
}

void waitForConnection(int ttl) {
	if (ttl == 0) {
		display.clearDisplay();
		display.display();
		display.println("WiFi failed");
		display.println("");
		display.println("Rebooting");
		display.display();
		delay(3000);
		ESP.restart();
		return;
	}
	int counter = 0;

	display.setTextSize(1);
	display.setTextColor(BLACK);
	display.setCursor(0, 0);
	display.clearDisplay();
	display.println("Connecting ...");
	display.drawRect(2, 20, display.width() - 4, 10, BLACK);
	display.display();

	while (WiFi.status() != WL_CONNECTED) {
		display.fillRect(2 + counter, 20, 1, 10, BLACK);
		display.display();
		counter++;
		if (counter > 80) {
			waitForConnection(ttl - 1);
			break;
		}
		delay(50);
	}

	display.clearDisplay();
	display.println("");
	display.println("");
	display.println("WiFi connected");
	display.display();
}

void downloadEmails() {
	tcpCon.connect(pop3Server, 110);
	String response = tcpCon.readStringUntil('\r');
	tcpCon.readBytes(buff, 1);
	if (response.startsWith("+OK")) {
		pop3Command(pop3User);
		String response = pop3Command(pop3Password);
		if (!response.equals("")) { // Empty string is returned while timeout is thrown.
			String messageCount = response.substring(4, response.indexOf(' ', 4));
			messageCount.trim();

			if (messageCount.equals("")) {
#ifdef DEBUG
				Serial.print("Failed to auth to the POP3 server. Probably invalid username or password.");
#endif
				displayFailure();
			}

			String command = "retr " + messageCount + "\r\n";
			command.getBytes(buff, command.length(), 0);

			sendCommand((char*)buff);
			readEmail();
		}
		else {
#ifdef DEBUG
			Serial.print("Failed to auth to the POP3 server. Probably invalid username or password. RAW response from the server is: ");
			Serial.println(response);
#endif
			displayFailure();
		}
		tcpCon.stopAll();
	}
	else {

#ifdef DEBUG
		Serial.println("Failed to estabilish connection with POP3 server.");
#endif
		// TODO: Connction failed.
	}
}

/*
	Parses probably the most usefull part from FROM header. Email format could be

	"Jožin Zbažin" <jozin@zbazin.cz>

	but could miss osme parts. For example 

	<jozin@zbazin.cz>

	is still valid.
*/
String processFrom(String originalFrom) {

#ifdef DEBUG
	Serial.print("Procesing from: ");
	Serial.println(originalFrom);
#endif
	String output;
	if (originalFrom.startsWith("<")) {
		output = originalFrom.substring(1, originalFrom.indexOf(">"));
	}
	else if (originalFrom.startsWith("\"")) {
		String temp = originalFrom.substring(1);
		output = temp.substring(0, temp.indexOf("\""));
	}
	else if (originalFrom.indexOf("<") != -1) {
		output = originalFrom.substring(0, originalFrom.indexOf("<"));
	}
	else {
		output = originalFrom;
	}
	output.trim();
	return decodeRfc2047(output);
}

/*
	Parse subject value as described in RFC 2047 https://www.ietf.org/rfc/rfc2047.txt

	For example: if original subject is =?utf-8?q?=C4=8Ce=C5=A1tina?= then output is Čeština
	For example: if original subject is =?UTF-8?B?xIxlxaF0aW5h?= then output is Čeština
*/
String decodeRfc2047(String originalSubject) {
	if (originalSubject.startsWith("=?")) {
		String temp = originalSubject.substring(2);
		temp = temp.substring(temp.indexOf("?") + 1); // skip encoding
		char format = temp.charAt(0);
		String data = temp.substring(2, temp.lastIndexOf("?="));
		String output;

		if (format == 'q' || format == 'Q') {
			while (data.indexOf("=") != -1) {
				String code = data.substring(data.indexOf("=") + 1, data.indexOf("=") + 3);
				char buff[3];
				code.toCharArray(buff, 3, 0);
				buff[2] = '\0';
				unsigned char number = (unsigned char)strtol(buff, NULL, 16);
				buff[0] = number;
				buff[1] = '\0';
				data.replace("=" + code, String(buff));
			}
			output = data;
		}
		else if (format == 'b' || format == 'B') {
			char decoded[301];
			char input[300];
			int toCopy = 300;

			memset(decoded, 0, 301);
			if ((data.length() + 1) < toCopy) {
				toCopy = data.length() + 1;
			}
			data.toCharArray(input, toCopy);
			b64_decode(decoded, input, toCopy);
			output = String(decoded);
		}
		else {
			output = "";
		}
		output.replace("_", " ");
		return output;
	}
	else {
		originalSubject.trim();
		originalSubject.replace("_", " ");
		return originalSubject;
	}
}

/*
	Removes special behaviour of some czech characters. It override input data and also returns it.

	For example: if data is Čeština then data would be Cestina
*/
String fixCzechCharacters(String data) {
	data.replace("Ě", "E");
	data.replace("Š", "S");
	data.replace("Č", "C");
	data.replace("Ř", "R");
	data.replace("Ž", "Z");
	data.replace("Ý", "Y");
	data.replace("Á", "A");
	data.replace("Í", "I");
	data.replace("É", "E");
	data.replace("Ů", "U");
	data.replace("Ú", "U");
	data.replace("Ď", "D");
	data.replace("Ť", "T");
	data.replace("Ň", "N");
	data.replace("ě", "e");
	data.replace("š", "s");
	data.replace("č", "c");
	data.replace("ř", "r");
	data.replace("ž", "z");
	data.replace("ý", "y");
	data.replace("á", "a");
	data.replace("í", "i");
	data.replace("é", "e");
	data.replace("ů", "u");
	data.replace("ú", "u");
	data.replace("ď", "d");
	data.replace("ť", "t");
	data.replace("ň", "n");
	return data;
}

/*
	Handles TCP connection to POP3 server and parses email from that.
*/
void readEmail() {

#ifdef DEBUG
	Serial.println("READING EMAIL");
#endif
	String line = waitForAndReadLine();
	line.trim();

	String from;
	String subject;
	String date;

	INIT_TIMEOUT;
	while (line[0] != '.' && CHECK_TIMEOUT_NOT_HAPPEND(10000)) { // timeout 10sec
		line = waitForAndReadLine();
		line.trim();

		if (line.startsWith("From: ")) {
			from = fixCzechCharacters(processFrom(line.substring(6)));
		}

		if (line.startsWith("Subject: ")) {
			subject = fixCzechCharacters(decodeRfc2047(line.substring(9)));
		}

		if (line.startsWith("Date: ")) {
			date = line.substring(6);
		}
	}
	CHECK_TIMEOUT_HAPPEND(10000, )

#ifdef DEBUG
	Serial.println("=========================================");
	Serial.println("DATE=" + date);
	Serial.println("FROM=" + from);
	Serial.println("SUBJECT=" + subject);
	Serial.println("=========================================");
#endif
	if (!lastDate.equals(date) || !lastFrom.equals(from) || !lastSubject.equals(subject)) {
		lastDate = date;
		lastFrom = from;
		lastSubject = subject;

		digitalWrite(pinLED, 1);
		showOnDisplay(from, subject, date);
	}

}

void showOnDisplay(String from, String subject, String date) {
	display.clearDisplay();
	display.setTextSize(1);
	display.setCursor(0, 0);
	display.println("Od:" + from);
	display.println("Predmet:" + subject);
	display.display();
}

#define LINE_BUFF_CAPACITY 100
char waitForAndReadLineBuff[LINE_BUFF_CAPACITY];

/*
	Reads one line from TCP stream to POP3 server. Returns it. Output does not contains \r\n.
*/
String waitForAndReadLine() {
	int buffOffset = 0;
	INIT_TIMEOUT;
	while (CHECK_TIMEOUT_NOT_HAPPEND(10000)) {
		INIT_TIMEOUT2;
		while (!tcpCon.available() && CHECK_TIMEOUT_NOT_HAPPEND2(10000)) {
			delay(1);
		}
		CHECK_TIMEOUT_HAPPEND2(10000, "")

		for (int i = 0; i < tcpCon.available() && buffOffset < LINE_BUFF_CAPACITY; i++) {
			char byte[1];
			tcpCon.readBytes(byte, 1);

			if (byte[0] == '\r') {
				waitForAndReadLineBuff[buffOffset] = '\0';
				return String(waitForAndReadLineBuff);
			}
			else {
				waitForAndReadLineBuff[buffOffset++] = byte[0];
			}
		}


		if (buffOffset == LINE_BUFF_CAPACITY) {
			// Buffer is full, waste all next bytes until \r

			char byte = '\0';
			while (byte != '\r') {
				INIT_TIMEOUT2;
				while (!tcpCon.available() && CHECK_TIMEOUT_NOT_HAPPEND2(10000)) {
					delay(1);
				}
				CHECK_TIMEOUT_HAPPEND2(10000, "");
				tcpCon.readBytes(&byte, 1);
			}

			return String(waitForAndReadLineBuff);
		}
	}
	CHECK_TIMEOUT_HAPPEND(10000, "");
}

/*
	Executes an POP3 command at the server. This function returns output from the server
*/
String pop3Command(char* command) {
	sendCommand(command);
	INIT_TIMEOUT;
	while (tcpCon.available() < 3 && CHECK_TIMEOUT_NOT_HAPPEND(10000)) {
		delay(1);
	}
	CHECK_TIMEOUT_HAPPEND(10000, "")
	int avail = tcpCon.available();

#ifdef DEBUG
	Serial.println(String(avail));
#endif
	tcpCon.readBytes(buff, avail);
	String response = String((char*)buff);
	return response;
}

/*
	@internal
	Executes an command on the POP3 server. This function only do required stuff to send the command to the server. It does not process output anyway.
*/
void sendCommand(char* command) {

#ifdef DEBUG
	Serial.print("CMD: ");
	Serial.println(command);
	Serial.println("WASTE: " + String(tcpCon.available()) + " bytes");
#endif
	INIT_TIMEOUT;
	while (tcpCon.available() && CHECK_TIMEOUT_NOT_HAPPEND(10000)) {
		tcpCon.read();
	}
	CHECK_TIMEOUT_HAPPEND(10000, );
	tcpCon.write(command);
	tcpCon.write("\r\n");
}

void loop() {
	downloadEmails();
	delay(60000);
}
